## Copyright (C) 1998-2001 by
## Steve Haynal
## Department of Electrical and Computer Engineering,
## University of California, Santa Barbara.
##
##  Permission to use, copy, modify and/or distribute this software
##  and its documentation for any purpose is hereby granted without
##  fee, subject to the following terms and conditions:
##
##  1.  The above copyright notice and this permission notice must
##  appear in all copies of the software and related documentation.
##
##  2.  This software may not be called "PYABSS" or "ABSS" if it has
##  been modified in any way, without the specific prior permission
##  of Andrew Seawright.
##
##  3.  THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
##  EXPRESSED, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
##  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
##
##  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA OR THE AUTHORS OF THIS
##  SOFTWARE BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
##  CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING
##  FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
##  POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF
##  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
##
##  notes:
##
##  The development of this software was supported through grants from
##  Intel Inc., the University of California Micro Program and
##  National Science Foundation
##                                            

##############################################################################################
## This file describes the MA class. It contains methods for composition and exploration.     
## ABSS dissertation chapters 4 and 5, with emphasis on chapter 4, are implmented here.               
##
## Steve Haynal, 1/15/01
##
## This file is organized as follows:
##     Composition and construciton related methods (Dissertation chapter 4)
##     High-level exploration methods (Dissertation chapter 5)
##     ROBDD ordering support methods
##     Dot visualization support methods
##     Witness printing support methods
##
## To Do:
##

from pyfa import *

## Random numbers are needed DFTightlyClosed
import whrandom
whrandom.seed(30,40,50)

class Model :
    ## State base partition and Smooth base partition may be specified
    def __init__(self,parent,stpart=0,smpart=0):
        self.stpart = stpart
        self.smpart = smpart
        self.parent = parent
        if (self.parent):
            self.parent.submodels.append(self)

    ## Determine and return dual of term
    def Dual(self,term):
        return self.vecDual.Compose(term)


################################################################################
## Composition and construction related methods

    ## Builds a composite-member's TR (data-phase transition relation)
    ## assumes self.transitions contains a list of transitions
    ## self.smth has been assigned
    ## Base method, may be overridden if desired
    def BuildTR(self):

        ## Base component lives in one partition
        self.TR = Trans(1)

        ## Transitions
        self.TR[0].rel = VecOr(self.transitions)
        
        ## Smoothing -- state objects identifying ROBDD variable assignment for present (P), next (N) and tags (T) 
        self.TR[0].pssmth[0] = self.smth.P
        self.TR[0].nssmth[0] = self.smth.N
        self.TR[0].tagsmth[0] = self.smth.T
        
        ## Maping -- maps are used to shift variables from next state to present state 
        self.TR.SetMapInfo(self.smth.P,self.smth.N,self.smth.T)

    ## Builds a composite-member's cTR (control-phase transition relation)
    ## Must be called after BuildTR
    ## Base method, may be overridden if desired
    def BuildcTR(self):

        ## Copy existing TR
        self.cTR = Trans(0,self.TR)

        ## Holding term (States just idle)
        hTR = self.TR.vecps.HoldTR(self.TR.vecns)

        ## Task bypass labels (Automatically determine bypass from taskstart definitions)
        self.bypass = ( ( ~self.taskstart[0][1].P & self.taskstart[0][1].N, ~self.taskstart[0][0].P & self.taskstart[0][0].N ), )

        ## Add bypass terms
        hTR = hTR + ( self.bypass[0][0] + self.bypass[0][1] )

        ## Substitue new relation
        self.cTR[0].rel = hTR

    ## Builds all composite state by specifying MA in self.submodels list
    def BuildCompositeState(self,base):

        self.base = base
        top = base
        for sm in self.submodels:
            sm.Specify(top)
            top = top + sm.width

        self.width = top - self.base

        ## Define sift groups after initial specification
        for sm in self.submodels:
            sm.BuildSiftGroup()

    ## Builds a composite data-phase TR
    def BuildCompositeTR(self,defaultpartitions=1):

        self.TR = Trans(defaultpartitions)

        self.TR.PreCompose()

        self.taskstart = [ [State(0), ], ]
        self.taskfinal = [ [State(0), ], ]
       
        for sm in self.submodels:
	    sm.BuildTR()
            self.TR.Add(sm.TR,sm.stpart,sm.smpart)

            # Flatten starting state
            for i in range(0,len(sm.taskstart)):
                for j in range(0, len(sm.taskstart[i])):
                    try:
                        self.taskstart[i][j] = self.taskstart[i][j] & sm.taskstart[i][j]
                    except IndexError:
                        self.taskstart[i].append(State(0))
                        self.taskstart[i][j] = self.taskstart[i][j] & sm.taskstart[i][j]
                        
            # Flatten final state
            for i in range(0,len(sm.taskfinal)):
                for j in range(0, len(sm.taskfinal[i])):
                    try:
                        self.taskfinal[i][j] = self.taskfinal[i][j] & sm.taskfinal[i][j]
                    except IndexError:
                        self.taskfinal[i].append(State(0))
                        self.taskfinal[i][j] = self.taskfinal[i][j] & sm.taskfinal[i][j]
                        

        self.TR.Compose()

    ## Builds a composite control-phase cTR
    def BuildCompositecTR(self,defaultpartitions=1):

        self.cTR = Trans(defaultpartitions)

        self.cTR.PreCompose()
       
        for sm in self.submodels:
	    sm.BuildcTR()
            self.cTR.Add(sm.cTR,sm.stpart,sm.smpart)

        self.cTR.Compose()

    ## Constructs a composite's commonly used data-structures
    ## Flips ps only, ignores guard ps bits.
    ## self.vecAll is all ROBDD variables under the manager
    ## self.vecDual is all ROBDD varables with non-control ps complemented
    ## self.outsame is ROBDD term to used force tag equal to ps. Is only built for outputs in outtags 
    def BuildVectors(self, partslice=(0,0), smthslice=(0,0)):

        ## Determine default partitions to apply
        if partslice[0] == partslice[1]: partslice = (0,len(self.TR.partitions))

        ## ROBDD vector of all variables under the manager
        self.vecAll = DdArray(-1)
        self.vecAll.Fill(self.base,1)

        ## Vec flip bitwise complements all ps variables EXCEPT control values
        ## Called with smthslice=(0,1) so as to exclude control values which by
        ## convention are in partition 1
        self.vecDual = DdArray(-1)
        self.vecDual.Fill(self.base,1)

        ## Determine all variables which must "flip" to compute a dual
        self.dual = mgr.One()
        for sm in self.submodels:
            self.dual = self.dual & sm.dual
                          
        ## Flip appropriate variables
        for i in range(0,len(self.vecDual)):
            t = ~self.vecDual[i]
            if (t & self.dual).Empty():
                self.vecDual[i] = t

        ## Atempt to build out same
        try:
            vec = DdArray(len(self.outtags))
            for ma in self.outtags:
                term = mgr.One()
                for i in range(0,len(ma.TR.vecps)):
                    term = term & ma.TR.vecps[i].Xnor(ma.TR.vectag[i])
                vec.Push(term)
            self.outsame = vec.And()
        except AttributeError:
            print "No outsame built"
            self.outsame = mgr.One()

    ## Creates a member self.rd or reverse object-to-name dictionary
    ## assignes self.name to all submodels
    ## Assumes all submodels have instantiated
    def AddNames(self):
        self.rd = {}
        items = self.__dict__.items()
        for nd in self.submodels:
            for (key, value) in items:
                if nd == value:
                    self.rd[nd] = key
        for sm in self.submodels:
            sm.name = self.rd[sm]

    ## Add outputs (produced information) to all MA in a composition
    def BuildOOutputs(self):
        try:
            i = 0
            for ilist in self.oinputs:
                for dt in ilist: 
                    dt.prd.ooutputs[dt.pi].append( dt )
                    dt.AddAccepter(self, i)
                i = i + 1
        except:
            print "No oinputs for",self.name,"in BuildOOutputs"

    ## Add additional control block information
    def AddCB(self,icb0,icb1,ncb0,ncb1):
        self.icb[0] = self.icb[0] | icb0
        self.icb[1] = self.icb[1] | icb1
        self.ncb[0] = self.ncb[0] | ncb0
        self.ncb[1] = self.ncb[1] | ncb1

    ## BuldTRcTRConstraints is the main routine for composite TR pruning for
    ## control examples. This routine builds correct dependency, operand resolution, memory care and task bypassing.
    ## Called for each submodel in a composition. setTR and setcTR are lists that collect constraint terms
    ## for the respective transition relation.
    def BuildTRcTRConstraints(self, setTR, setcTR):

        ## Build dependencies -- assumes single output to input, using renaming machines
        i = 0
        ## for each input of the composite-member MA
        for ilist in self.oinputs:

            ## genable collects operand resolution dependency constraints in both senses
            genable = [mgr.Zero(), mgr.Zero()]

            term0 = mgr.Zero()
            term1 = mgr.Zero()

            ## If input has multiple satisfying sources, then it is a join
            ## Control is assumed so no alternatives are currently allowed
            if len(ilist) > 1:

                ## gcb collects applicable control blocks and checks for control overlap
                ## among dependencies at an operand resolution
                gcb = [mgr.Zero(), mgr.Zero()]
                
                ## For each dependency tuple at this operand resolution
                for dt in ilist:
                    
                    ## Sanity checks on dt structure
                    if not (dt.acp == self):
                        print "WARNING: DT accepter is",dt.acp.name,"but built for",self.name
                    if not (i == dt.ai):
                        print "WARNING: DT accepter index is",dt.ai,"but built for",i

                    ## Compute applicable control block. Intersection of producer's cb, accepter's cb and any special conditions
                    guard0 = self.icb[0].P & dt.prd.icb[0].P & dt.pcb[0].P
                    guard1 = self.icb[1].P & dt.prd.icb[1].P & dt.pcb[1].P

                    ## Applicable control block should not be empty, otherwise why specify this dependency
                    if guard0.Empty() or guard1.Empty():
                        print "WARNING: Guard is empty for operand resolution from",dt.prd.name,"to",self.name

                    ## compute previous terms that are inside and outside of the applicable control block
                    inside0 = guard0 & gcb[0]
                    outside0 = guard0 & ~gcb[0] 

                    inside1 = guard1 & gcb[1]
                    outside1 = guard1 & ~gcb[1]

                    ## Further constrain already covered control-blocks
                    if not inside0.Empty():
                        print "WARNING: Found overlapping control for",self.name,"input",i
                        ## Determine portion to constrain
                        toconstrain = genable[0] & inside0
 
                        ## remove toconstrain
                        genable[0] = genable[0] - toconstrain
 
                        ## add inside0 back
                        constrained = toconstrain & dt.prd.known[dt.pi][dt.si].P
                        genable[0] = genable[0] + constrained

                    if not inside1.Empty():

                        ## Determine portion to constrain
                        toconstrain = genable[1] & inside1
 
                        ## remove toconstrain
                        genable[1] = genable[1] - toconstrain
 
                        ## add constrained back
                        constrained = toconstrain & dt.prd.known[dt.pi][abs(dt.si-1)].P
                        genable[1] = genable[1] + constrained
                        
        
                    ## Include possible resolution for previously uncovered contgrol cases
                    genable[0] = genable[0] + (outside0 & dt.prd.known[dt.pi][dt.si].P)
                    genable[1] = genable[1] + (outside1 & dt.prd.known[dt.pi][abs(dt.si-1)].P)

                    ## gcb does not include producer known information
                    gcb[0] = gcb[0] + outside0
                    gcb[1] = gcb[1] + outside1

                    ## Update memory care condition
                    ## mc0 in words, "Control has resolved and is appicable OR control has NOT yet resolved in this sense"
                    mc0 = guard0 + ~( dt.prd.icb[dt.si].P + dt.prd.ncb[dt.si].P )
                    mc1 = guard1 + ~( dt.prd.icb[abs(dt.si-1)].P + dt.prd.ncb[abs(dt.si-1)].P )
                    ## Producer's memcare in words, "Previous memory care OR producer is known AND memory care is true AND child has NOT accepted"
                    dt.prd.memcare[dt.pi][dt.si] = dt.prd.memcare[dt.pi][dt.si] + (dt.prd.stored[dt.pi][dt.si].P & mc0 & ~self.accepted[i][0].P)
                    dt.prd.memcare[dt.pi][abs(dt.si-1)] = dt.prd.memcare[dt.pi][abs(dt.si-1)] + (dt.prd.stored[dt.pi][abs(dt.si-1)].P & mc1 & ~self.accepted[i][1].P)

            ## Standard operand dependency tuple with a sink only paired with 1 source
            elif len(ilist)==1:
                ## Not a join hence ilist only contains one element at index 0
                dt = ilist[0]

                ## Sanity checks on dt structure
                if not (dt.acp == self):
                    print "WARNING: DT accepter is",dt.acp.name,"but built for",self.name
                if not (i == dt.ai):
                    print "WARNING: DT accepter index is",dt.ai,"but built for",i

                ## Producer's control block must always contain accepter's control block
                if not self.icb[0].P.Leq(dt.prd.icb[0].P):
                    print "WARNING: Producer's control block does not contain accepter's control block"
    
                term0 = self.required[i][0] >> dt.prd.known[dt.pi][dt.si].P
                term1 = self.required[i][1] >> dt.prd.known[dt.pi][abs(dt.si-1)].P

                ## Update memory care condition
                dt.prd.memcare[dt.pi][dt.si] = dt.prd.memcare[dt.pi][dt.si] + (dt.prd.known[dt.pi][dt.si].P & ~self.accepted[i][0].P)
                dt.prd.memcare[dt.pi][abs(dt.si-1)] = dt.prd.memcare[dt.pi][abs(dt.si-1)] + (dt.prd.known[dt.pi][abs(dt.si-1)].P & ~self.accepted[i][1].P)
 
            else:
                print "WARNING: MA",self.name,"has no inputs specified for input",i
                term0 = mgr.One()
                term1 = mgr.One()
    
            if not genable[0].Empty():
                print "Operand resolution point for",self.name,"input",i
                if not term0.Empty():
                    print "WARNING: Mixing regular and guarded dependencies for", self.name,"input",i
                term0 = self.required[i][0] >> genable[0]
                term1 = self.required[i][1] >> genable[1]

            ## Save dependency implications
            setTR.append(term0 & term1)

            i = i + 1

        ## Task bypass if sense is known in current iteration sense
        ## AND self is control-obviated
        ## AND self is not in the other iteration taskstart
        term0 = self.parent.sense.known[0][0].N & self.ncb[0].N & ~self.taskstart[0][1].P
        term1 = self.parent.sense.known[0][1].N & self.ncb[1].N & ~self.taskstart[0][0].P
        
        if (term0.Empty() or term1.Empty()):
            print "WARNING: Term0 or term1 in a bypass construction is empty for",self.name

        ## term true AND must go to bypass state OR
        ## term not true AND prohibit bypass
        imp0 = (term0 & self.taskstart[0][1].N) + ( ~term0 & ~self.bypass[0][0] )
        imp1 = (term1 & self.taskstart[0][0].N) + ( ~term1 & ~self.bypass[0][1] )

        ## Collect bypass terms
        setcTR.append(imp0 & imp1)

    ## Build Memory routines
    ## For data-flow construction
    ## Memory care is producer stores result yet child has not accepted
    def BuildMemory(self,vali,grp2):
        minl = min(len(self.stored[vali]),len(grp2))
        for i in range(0,minl):
            self.memcare[vali][i] = self.memcare[vali][i] + (self.stored[vali][i].P & ~grp2[i].P)

    def BuildIMemory(self,vali,grp2):
        minl = min(len(self.stored[vali]),len(grp2))
        for i in range(0,minl):
            self.memcare[vali][i] = self.memcare[vali][i] + (self.stored[vali][i].P & ~grp2[(minl-1)-i].P)

###########################################################################
## Top-level exploration methods for data-flow examples

    def DFSchedule(self,looppipe=1):

        start = time.clock()

        ## if looppipe is 1, then create a behavior cut for looppipelining
        if looppipe==1:
            # Create Sbc
            ## FIXME: Match pin sense with other scheduling routines
            Sbc = self.pin.known[0][0].N & ~self.pin.known[0][0].P
            Sbc = self.TR.Image(Sbc)
            eterm = self.pin.known[0][1].P # early termination check
        ## use starting state -- no loop overlap
        else:
            Sbc = self.taskstart[0][0].P
            eterm = self.taskfinal[0][0].P

        # Tag Sbc
        Sbc = Sbc & self.outsame
     
        # Create path set
        ps = PathSet(self,Sbc)

        # Forward exploration
        # Loop pin in opposite sense is early termination indicator
        # Start at time-step set 0
        ps.ForwardExploration(0,eterm)

        while 1:

            # Find closed ps if possible
            closedps = ps.Closure()

            # If closure was successful
            if closedps:
                closedps.FindTerms(closedps.rfs)
                tightlyclosedps = self.DFTightlyClosed(closedps)
                if tightlyclosedps:
                    singleclosedps = self.DFSingleSchedule(tightlyclosedps)
                    if singleclosedps:
                        print
                        break

            ps.ForwardExploration(-1,eterm)

        # Closed schedule found
        print "Schedule with throughput of",singleclosedps.last(),"found in",time.clock()-start,"seconds"
        print 'Peak nodes',mgr.ReadPeakNodeCount()
        return singleclosedps

    def DFTightlyClosed(self,exset):
        start = time.clock()
        rand = whrandom.random()
        # copy working set
        wset = PathSet(0,0,0,0,exset)

        # Smooth all tags, including outtags
        wset.SmoothTimeSteps(wset.TR.tag) ## smooth tags
        wset.fs = wset.fs.ExistAbstract(wset.TR.tag)
        wset.rfs = wset.rfs.ExistAbstract(wset.TR.tag)

        jlist = range(0,len(wset.TR.vecps))

        ## Remove current tags
        otagterm = mgr.One()
        for otag in self.outtags:
            otagterm = otagterm & otag.smth.P

        ## Remove io protocol if present
        try:
            otagterm = otagterm & self.io.smth.P
        except AttributeError:
            print "No io protocol"
            pass
        
        ilist = []
        for j in jlist:
            if not (otagterm.Leq(wset.TR.vecps[j])):
                ilist.append(j)

        while (len(ilist) > 0):
            tagterm = mgr.One()
            etagsmth = mgr.One()
        
            for j in range(0,self.tagdepth):
                if len(ilist) == 0:
                    break

                ## Assign a state bit
                randi = whrandom.choice(ilist)

                ttt = ( wset.TR.vecps[randi].Xnor(wset.TR.vectag[randi]) )
                tagterm = tagterm & ttt
                etagsmth = etagsmth & wset.TR.vectag[randi]

                ilist.remove(randi)
             
                if len(ilist) == 0:
                    break

            ## tag wset
            wset[0] = wset[0] & tagterm
                      
            ## attempt closure
            ## TightlyClosed does not accumulate early terminations as we are looking
            ## for the special case of single iteration latency == control steps
            result = wset.Closure(0)
            result.rfs = result[result.last()]
            
            if result == 0:
                print "Single Sets Failure", time.clock()-start
                return result
            wset = result
            wset.SmoothTimeSteps(etagsmth) ## smooth tags

        print 'Single Sets', time.clock()-start
                
        return wset

    def DFSingleSchedule(self,wset):

        start = time.clock()

        pool = wset[wset.last()]
 
        i = 1
        while not pool.Empty():

            attempt = self.vecAll.PickOneMinterm(pool)
            attempt = self.TR.SmoothNS(attempt)

            try:
                attempt = attempt.ExistAbstract(self.io.smth.P)
            except AttributeError:
                print "No IO protocol in single schedule"
                pass

            wset.rfs = attempt
            result = wset.Closure()

            if result:
                print 'Single' + repr(i), time.clock()-start
                return result
                
            pool = pool - attempt
            print "SS" + repr(i),
            i = i+1
 
        print 'No Single Schedule in',time.clock()-start,'seconds at step',i+1
        return 0


###########################################################################
## Top-level exploration for control examples


    ## Check if term covers all control cases
    def AllPathsCheck(self,term):
        
        term = self.TR.SmoothPS(term,(0,0),(0,1))
        term = self.TR.SmoothTS(term,(0,0),(0,1))
        if not (term == mgr.One()):
            print "All Paths Failure"
            return 0
        return 1

    ## Viability prune for models with both TR and cTR
    def DualReachTRcTR(self,sslist):

        nrs = mgr.Zero()
        for arg in sslist:
	    nrs = nrs | arg.P
            
        rs = mgr.Zero()

        i = 1
        while not (nrs == rs):
            rs = nrs
            print "TR",
            nrs1 = self.TR.Image(rs)
            print "cTR",
            nrs2 = self.cTR.Image(nrs1)
            print "GFP" + repr(i),
            nrs = nrs1 + nrs2
            i = i + 1

        print
        
        return rs
    
    ## For behavior with control, All non-stalling schedules on all paths, not just minimum latency schedules. 
    def CSchedule(self):

        ## Create Sbc 
        Sbc = self.pin.known[0][1].P & ~self.pin.known[0][1].N
        Sbc = self.TR.Intersect(Sbc)
        Sbc = self.TR.SmoothNS(Sbc)

        Sbc_ = self.Dual(Sbc)
        ## Check before Sbc is tagged
        if not Sbc.Leq(~Sbc_):
            print "WARNING: Sbc and Sbc~ contain common states"

        if not self.AllPathsCheck(Sbc):
            print "WARNING: Sbc does not cover all control cases"

        ## Tag Sbc
        Sbc = Sbc & self.outsame

        ## TR and cTR are pruned so as not to leave Sbc~
        ## Once Sbc~ is reached, you can not leave, only transit to another Sbc~ state
        ## FIXME: This needs some careful thought. Can you ever reach Sbc~ leave and then return in a valid schedule? I don't think so.
        Sbc_ = self.Dual(Sbc)
        self.TR.PruneT( ~(Sbc_ & ~(Sbc_.VarMap())))
        self.cTR.PruneT(~ (Sbc_ & ~(Sbc_.VarMap())))

        ## Create path set
        ps = PathSet(self,Sbc)

        ## Forward exploration
        ps.FullForwardExploration()

        return ps

 
    ## Fixed-point over all control cases
    ## Universal abstraction must be done separately per control-case
    def Validate(self,term):

        ## NS guards only, PS everything
        j = len(self.forks)
        while j > 0:
            for frk in self.forks:
                toprune = frk.resolve.P & term
                therest = ~frk.resolve.P & term
                pruned = toprune.UnivAbstract(frk.univabs.N)

                ## Validation pruned terms
                if pruned!=toprune:
                    term = pruned + therest
                    ## Reset j, must reach fixed-point for all forks, pruned terms may break earlier validation passes
                    j = len(self.forks)
                    print "VP",

                ## Validation did not prune terms
                else:
                    print "VNP",
                    
                j = j-1
                if j==0:
                    break
                if term.Empty():
                    print "WARNING: Validation pruned term to empty!"
                    j=0
                    break
                
        return term

    ## This routine prioritizes (maximizes throughput) a list of control cases (cplist)
    ## Exset is a validated pathset
    ## Priority follows order of the list
    def PrioritizeControlCases(self, exset, cplist):

        termlist = []
        ## working copy 1
        wset1 = PathSet(0,0,0,0,exset)

        i = 0
        ## For each priority term
        for cp in cplist:

            i = i + 1
            print "Attempting control case",i

            ## create prune term and task final states reached for this control case
            pterm = wset1.rfs & cp

            ## find earliest set where particular control cases terminate -- Start from beginning as there may be gaps from end
            allterm = mgr.Zero() ## Accumulates all terminations
            for j in range(0,wset1.last()+1,2):
                ## New terminations
                newterm = wset1[j] & pterm

                ## Acucmulate all terminations
                allterm = allterm + newterm

                ## Check for containment
                tcp = wset1.TRlist[0].SmoothPS(allterm,(0,0),(0,1))
                tcp = wset1.TRlist[0].SmoothTS(tcp,(0,0),(0,1))

                if (cp == tcp):
                    print "Found termination for control case",i,"in time step set",j
                                       
                    ## working copy 2
                    wset2 = PathSet(0,0,0,0,wset1)

                    ## Exclude any future activity for the prioritized control case
                    for k in range(j+1,len(wset2.tslist) ):
                        wset2.tslist[k] = wset2.tslist[k] & ~cp

                    ## Update rfs to only include early terminations for the prioritized control case
                    wset2.rfs = (wset2.rfs & ~cp) + allterm
       
                    ## Try to validate and close
                    result = wset2.ValidatedClosure()
                      
                    if result:
                        print "Verified closed repeatable termination for control case",i,"in time step set",j
                        termlist.append(j)
                        wset1 = result
                        break

                else:
                    print "No termination for control case",i,"in time step set",j

        print "Termination list is",termlist

        ## Save for later printing
        wset1.termlist1 = termlist

        return wset1

    ## This method finds best possible latencies irregardless of validation
    ## This is called after validation and searches for the shortest possible sequence for the listed control cases
    ## Such shortest sequence need only exist and it may only be reachable after certain (not all) previous control cases
    def BestControlCases(self, exset, cplist):

        termlist = []

        i = 0
        ## For each control case
        for cp in cplist:

            i = i + 1
            print "Attempting control case",i

            ## create prune term and task final states reached for this control case
            pterm = exset.rfs & cp

            ## find earliest set where particular control cases terminate -- Start from beginning as there may be gaps from end
            allterm = mgr.Zero() ## Accumulates all terminations
            for j in range(0,exset.last()+1,2):
                ## New terminations
                newterm = exset[j] & pterm

                ## Acucmulate all terminations
                allterm = allterm + newterm

                ## Check for containment
                tcp = exset.TRlist[0].SmoothPS(allterm,(0,0),(0,1))
                tcp = exset.TRlist[0].SmoothTS(tcp,(0,0),(0,1))

                ## Containment (equality) of only control caes
                if (cp == tcp):
                    print "Found termination for control case",i,"in time step set",j
                                       
                    ## copy set
                    wset2 = PathSet(0,0,0,0,exset)

                    ## Truncate wset2
                    wset2.Truncate(j)

                    ## Update rfs
                    wset2.rfs = allterm
       
                    ## Check if closure is possible -- no validation
                    result = wset2.ValidatedClosure(0)
                      
                    if result:
                        ## Check for containment
                        tcp = wset2.TRlist[0].SmoothPS(result[0],(0,0),(0,1))
                        tcp = wset2.TRlist[0].SmoothTS(tcp,(0,0),(0,1))

                        if not (cp == tcp):
                            print "WARNING: Not all control cases are included after closure"

                        print "Verified closed repeatable termination for control case",i,"in time step set",j
                        termlist.append(j)
                        break

                else:
                    print "No termination for control case",i,"in time step set",j

        print "Termination list is",termlist

        exset.termlist2 = termlist
        

################################################################################
## ROBDD Order related methods

    ## DumpOrder creates and prints a list describing the current ROBDD order
    ## position i of a list corresponds to ROBDD order position
    ## the element at i in a list is the ROBDD index at that position
    def DumpOrder(self):
        size = mgr.ReadSize()
        orderlist = []
        
        for i in range(0,size):
            orderlist.append(mgr.ReadInvPerm(i))

        print orderlist
        return orderlist

    ## ImportOrder reorders the ROBDD as specified in orderlist
    ## position i of a list corresponds to ROBDD order position
    ## the element at i in a list is the ROBDD index at that position
    def ImportOrder(self,orderlist):

        ivec = pycudd.IntArray(len(orderlist))

        i = 0
        for item in orderlist:
            ivec[i] = item
            i = i + 1

        mgr.ShuffleHeap(ivec)
 
    ## Groups MA ROBDD variables together for ROBDD sift reordering
    ## Called after the composition is constructed
    ## Mtrvalue of 0x04 specifies ROBDD sift reordering 
    def BuildSiftGroup(self,mtrvalue=0x04):
        mgr.MakeTreeNode(self.base,self.width,mtrvalue)


    ## Forces a ROBDD reordering to occur
    ## Mtrvalue of 0x04 specifies ROBDD sift reordering
    ## See CUDD documentation for valid reordering types
    def Reorder(self,mtrvalue=0x04,minsize=0):
    	mgr.ReduceHeap(mtrvalue,minsize)
        

################################################################################
## DOT visualization related methods
## See http://www.research.att.com/sw/tools/graphviz/ for details regarding
## DOT and graphviz

    ## Creates DOT format entry for a graph node representing a composite-member task
    ## outfile is an open file handle
    def DotNodeLine(self,outfile):

        str = '\t' + self.name + '[label = "' + self.name + '|{'

        width = max(len(self.oinputs),len(self.ooutputs))

        ## Dump information input (information this task accepts)
        istr = '{ '
        for i in range(0,len(self.oinputs)):
            if i > 0:
                istr = istr + '|'
            istr = istr + '<i' + repr(i) + '> i' + repr(i)
        istr = istr + '}'

        ## Dump information output (information this task produces)
        ostr = '{'
        for i in range(0,len(self.ooutputs)):
            if i > 0:
                ostr = ostr + '|'
            ostr = ostr + '<o' + repr(i) + '> o' + repr(i)
        ostr = ostr + '}'

        str = str + istr + '|' + ostr + '}'
        str = str + '"];\n'
        outfile.write(str)

    ## Creates DOT format entry for a graph edges representing information acceptance
    ## outfile is an open file handle
    def DotEdgeLines(self,outfile):
        for ilist in self.oinputs:
            for dt in ilist:
                dt.DotEdgeLine(outfile)

    ## Top-level method to output a DOT format graph of a composite task
    ## outfile is an open file handle
    def DumpDot(self,outfile):
        outfile.write("digraph dfg {\n\n")
        outfile.write("\tratio=auto\n")
        outfile.write('\tpage="8,10.5"\n')
        outfile.write("\tnode [shape = record, height=0.1, fontsize = 8, fontname = Helvetica, style = solid ];\n\n")

        for sm in self.submodels:
            sm.DotNodeLine(outfile)

        for sm in self.submodels:
            sm.DotEdgeLines(outfile)

        outfile.write("\n\n}\n")


#####################################################################
## Schedule printing methods
            
    ## Print activity in a single transition term (ttr)
    def PrintTTR(self,ttr):

        ## Print transition information (function unit activity, res busy, etc.)
        for sm in self.submodels:
            sm.PrintRequired(ttr)
        print

        ## Print memory information 
        for sm in self.submodels:
            sm.PrintMemory(ttr)
        print

    ## Prints required information for a submodel
    def PrintRequired(self,term):

        # For all inputs
        cnt = 0
        # For all expected input
        for inpg in self.required:
            sense = 0
            # For all senses
            for inp in inpg:
                if not (inp & term).Empty():
                    # Print input event
                    output = self.name + "_in_" + repr(cnt) + "_" + repr(sense)
                    print output,
                sense = sense + 1
            cnt = cnt + 1

    ## Prints required store information for a submodel
    def PrintMemory(self,term):

        # For all possible registered results
        cnt = 0
        for cmg in self.memcare:
            sense = 0
            # For all senses
            for cm in cmg:                        
                if not (cm & term).Empty():
                    output = self.name + "_reg_" + repr(cnt) + "_" + repr(sense)
                    print output,
                sense = sense + 1
            cnt = cnt + 1

    ## Assumes iterates and control
    ## Prints only control and input 
    def PrintTTR2(self,ttr):

        i = 0
        for sm in self.submodels:
            print "Iterate",i,
            for ssm in sm.forks:
                ssm.PrintCB(ttr)
            
            for ssm in sm.submodels:
                ssm.PrintRequired(ttr)
            print
            i = i + 1

    ## Prints the current control block
    def PrintCB(self,term):

        c = 0
        for case in self.case:
            if not (case.P.Leq(~term)):
                print self.name + "=" + repr(c),
                break
            c = c + 1
            




            



















