## Copyright (C) 1998-2001 by
## Steve Haynal
## Department of Electrical and Computer Engineering,
## University of California, Santa Barbara.
##
##  Permission to use, copy, modify and/or distribute this software
##  and its documentation for any purpose is hereby granted without
##  fee, subject to the following terms and conditions:
##
##  1.  The above copyright notice and this permission notice must
##  appear in all copies of the software and related documentation.
##
##  2.  This software may not be called "PYABSS" or "ABSS" if it has
##  been modified in any way, without the specific prior permission
##  of Andrew Seawright.
##
##  3.  THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
##  EXPRESSED, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
##  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
##
##  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA OR THE AUTHORS OF THIS
##  SOFTWARE BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
##  CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING
##  FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
##  POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF
##  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
##
##  notes:
##
##  The development of this software was supported through grants from
##  Intel Inc., the University of California Micro Program and
##  National Science Foundation
##                                            

###################################################################################
## This file contains MA specifications.
## Refer to the ABSS dissertation chapter 3 for more information.
## The model COp1 below contains detailed comments. Use it as a template for your own MA specifications.
## Also of interest are AOp1 (acyclic model), Case and CaseExtend (multivalue MA)
##
## TO DO:
##   Some cyclic models may label more state known. COp2p, etc.
##   Both even and odd senses are manually specified. The symmetric half could be produced automatically.

from ma import *
from math import log

## Basic single time step cyclic unit
class COp1(Model):

    def Specify(self,base):

        ## State instantiation
        ## self.base is the base index in the composite ROBDD where this MA is instantiated
        ## self.width is the number of state bits times 3 for (ps, ns, tag)  to encode this MA 
        self.base = base
        self.width = 3
    
        ## States are listed and encoded explicitly with ASCII (0, 1, -)
        ## Empty state list
        set = []

        ## The names, such as S0 and S1, are an arbitrary choice
        S0 = State(self.base,'0',set)
        S1 = State(self.base,'1',set)

        ## Preserve set of states as self.states
        self.states = set

        ## All MA need to identify a state with all 1 encoding
        ## This state is used to build smoothing and mapping ROBDD arrays
        self.smth = S1

        ## ROBDD identifying what bits must be flipped to compute a present state dual
        self.dual = self.smth.P

        ## Transition instantiation
        ## Empty transition list
        set = []

        ## Add transitions
        ## One predecessor, one sucessor: Transition(set,S0,S0)
        ## One predecessor, several sucessors: Transition(set,S0,(S0,S1))
        S0_S0 = Transition(set,S0,S0)
        S0_S1 = Transition(set,S0,S1)
        S1_S1 = Transition(set,S1,S1)
        S1_S0 = Transition(set,S1,S0)

        ## Preserve set of transitions as self.transitions
        self.transitions = set

        ## Label states and transitions as in Table 3.1
        ## All labels are by nested tuples:
        ##   inner order is sense 0 through n (Only two are currently used: 0=even 1=odd)
        ##   outer order is instance for multiple instances (i.e. several output operand instances may be known)
        ##
        ##   Examples:
        ##     self.known = ( (S1, S0), (S1, S0) )
        ##                          ^ Output operand 0, sense 1 (odd)
        ##     self.known = ( (S1, S0), (S1, S0) )
        ##                                ^ Output operand 1, sense 0 (even)       

        ## State labels
        ## known label, complement is unknown
        ## two iteration senses, maximum two output operands (If desired, only one output operand may be used)
        self.known = ( (S1, S0), (S1, S0) )

        ## stored label
        ## two iteration senses, maximum two output operands
        self.stored = ( (S1, S0), (S1, S0) )

        ## accepted label
        ## two iteration senses, maximum three input operands
        self.accepted = ( (S1, S0), (S1, S0), (S1, S0) )

        ## taskstart label
        self.taskstart = ( (S0, S1), )

        ## taskfinal label
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        ## required label
        ## two iteration senses, maximum three input operands
        self.required = ( (S0_S1, S1_S0), (S0_S1, S1_S0), (S0_S1, S1_S0) )

        ## produced label
        ## two iteration senses, maximum two output operands
        self.produced = ( (S0_S1, S1_S0), (S0_S1, S1_S0) )

        ## forget label
        ## two iteration senses, maximum two output operands
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )

        ## busy label
        ## two iteration senses, maximum one resource use
        self.busy = ( (S0_S1, S1_S0), )

        


        ## The following information is created dynamically as it only has meaining when in a composition

        ## memory care conditions
        ## two iteration senses, maximum two output operands
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        
        ## List of dependency edges that this MA satisfies
        ## Organized by produced operands, mximum two output operands
        ## oouputs stands for operand outputs
        self.ooutputs = ( [], [] )

        ## self.oinputs (operand inputs) is specified within the composite specification

        ## Control block assignments
        ##   icb is in control block
        ##   ncb is not in control block
        ##   at first, don't care state place holders are used
        ##   order is by sense
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]


## Basic acyclic single time-step unit
## Acyclic models may be specified as follows with only one sense or cyclic
## models may be used in acyclic compositions and the second sense is ignored
## Maximum two inputs
## Maximum one output
class AOp1(Model):

    def Specify(self,base):

        self.base = base
        self.width = 3
    
        ## States 
        set = []
        S0 = State(self.base,'0',set)
        S1 = State(self.base,'1',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
       
        ## Transitions
        set = []
        S0_S0 = Transition(set,S0,S0)
        S0_S1 = Transition(set,S0,S1)
        S1_S1 = Transition(set,S1,S1)

        ## Preserve set of transitions as self.transitions
        self.transitions = set

        ## State labels
        self.known = ( (S1, ), )
        self.stored = ( (S1, ), )
        self.accepted = ( (S1, ), (S1, ) )
        self.taskstart = ( (S0, ), )
        self.taskfinal = ( (S1, ), )

        ## Transition labels
        self.required = ( (S0_S1, ), (S0_S1, ) )
        self.produced = ( (S0_S1, ), )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, ), )
        self.busy = ( (S0_S1, ), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), ], ]
        self.ooutputs = ( [], )
        self.icb = [State(0), ]
        self.ncb = [State(0), ]

## Two time-step pipelined acylic MA
class AOp2p(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P

        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S1_S1 =  Transition(set,S1,S1)
        self.transitions = set

        ## State labels
        self.known = ( (S1, ), (S1, ) )
        self.stored = ( (S1, ), (S1, ) )
        self.accepted = ( (S0p | S1, ), (S0p | S1, ), (S0p | S1, ) ) 
        self.taskstart = ( (S0, ), )
        self.taskfinal = ( (S1, ), )

        ## Transition labels
        self.required = ( (S0_S0p, ), (S0_S0p, ), (S0_S0p, ) )
        self.produced = ( (S0p_S1, ), (S0p_S1, ) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, ), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, ) )
        self.busy = ( (S0_S0p, ), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), ], [mgr.Zero(), ] ]
        self.ooutputs = ( [], )
        self.icb = [State(0), ]
        self.ncb = [State(0), ]


## Two time-step pipelined cyclic unit
## Maximum three inputs
## Maximum two outputs
class COp2p(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        S1p = State(self.base,'10',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P

        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S0 = Transition(set,S1p,S0)
        self.transitions = set

        ## State labels
        self.known = ( (S1, S0), (S1, S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p) ) 
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0_S0p, S1_S1p) )
        self.produced = ( (S0p_S1, S1p_S0), (S0p_S1, S1p_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p, S1_S1p), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Two time-step pipelined cyclic unit
## Maximum three inputs (third input occurs after first two)
## Maximum two outputs (first output occurs after single time-step)
class COp2ps(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        S1p = State(self.base,'10',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S0 = Transition(set,S1p,S0)
        self.transitions = set

        ## State labels
        self.known = ( (S0p | S1, S1p | S0), (S1, S0)  )
        self.stored = ( (S0p | S1, S1p | S0), (S1, S0) )
        self.accepted = ( (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p), (S1, S0) ) 
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0p_S1, S1p_S0) )
        self.produced = ( (S0_S0p, S1_S1p), (S0p_S1, S1p_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p, S1_S1p), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Two time-step nonpipelined cyclic unit
## Maximum three inputs
## Maximum two outputs
class COp2(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        S1p = State(self.base,'10',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S0 = Transition(set,S1p,S0)
        self.transitions = set

        ## State labels
        self.known = ( (S1, S0), (S1, S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p) ) 
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0_S0p, S1_S1p) )
        self.produced = ( (S0p_S1, S1p_S0), (S0p_S1, S1p_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p | S0p_S1, S1_S1p | S1p_S0), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Two time-step nonpipelined cyclic unit
## Maximum three inputs (third input occurs after first two)
## Maximum two outputs (first output occurs after single time-step)
class COp2s(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        S1p = State(self.base,'10',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S0 = Transition(set,S1p,S0)
        self.transitions = set

        ## State labels
        self.known = ( (S0p | S1, S1p | S0), (S1, S0)  )
        self.stored = ( (S0p | S1, S1p | S0), (S1, S0) )
        self.accepted = ( (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p), (S1, S0) ) 
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0p_S1, S1p_S0) )
        self.produced = ( (S0_S0p, S1_S1p), (S0p_S1, S1p_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p | S0p_S1, S1_S1p | S1p_S0), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Two time-step IO protocol machine
## Maximum three inputs
## Maximum two outputs
class COpRWIO(Model):

    def Specify(self,base):

        self.base = base
        self.width = 6

        ## States
        set = []
        S0 = State(self.base,'00',set)
        S0p = State(self.base,'01',set)
        S1 = State(self.base,'11',set)
        S1p = State(self.base,'10',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S1 = Transition(set,S0p,S1)
        S0p_S0p = Transition(set,S0p,S0p)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S0 = Transition(set,S1p,S0)
        S1p_S1p = Transition(set,S1p,S1p)
        self.transitions = set

        ## State labels
        self.known = ( (S1, S0), (S1, S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p), (S0p | S1, S0 | S1p) ) 
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )

        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0_S0p, S1_S1p) )
        self.produced = ( (S0p_S1, S1p_S0), (S0p_S1, S1p_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p | S0p_S1, S1_S1p | S1p_S0), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]
        
## Three time-step pipelined cyclic unit
## Maximum three inputs
## Maximum two outputs
class COp3p(Model):

    def Specify(self,base):

        self.base = base
        self.width = 9

        ## States
        set = []
        S0 = State(self.base,'000',set)
        S0p = State(self.base,'001',set)
        S0pp = State(self.base,'011',set)
        S1 = State(self.base,'111',set)
        S1p = State(self.base,'110',set)
        S1pp = State(self.base,'100',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S0pp = Transition(set,S0p,S0pp)
        S0pp_S1 = Transition(set,S0pp,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S1pp = Transition(set,S1p,S1pp)
        S1pp_S0 = Transition(set,S1pp,S0)
        self.transitions = set
 
        ## State labels
        self.known = ( (S1, S0), (S0p | S0pp | S1, S1p | S1pp | S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S0p | S0pp | S1, S1p | S1pp | S0), (S0p | S0pp | S1, S1p | S1pp | S0) )
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )
        
        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0_S0p, S1_S1p) )
        self.produced = ( (S0pp_S1, S1pp_S0), (S0_S0p, S1_S1p) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        ##self.busy = ( (S0_S0p, S1_S1p), )
	self.busy = ( (S0_S0p | S0pp_S1, S1_S1p | S1pp_S0), )
	## My dissertation modeled memory port busy during time-steps 1 and 3

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Three time-step nonpipelined cyclic unit
## Maximum three inputs
## Maximum two outputs
class COp3(Model):

    def Specify(self,base):

        self.base = base
        self.width = 9

        ## States
        set = []
        S0 = State(self.base,'000',set)
        S0p = State(self.base,'001',set)
        S0pp = State(self.base,'011',set)
        S1 = State(self.base,'111',set)
        S1p = State(self.base,'110',set)
        S1pp = State(self.base,'100',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S0pp = Transition(set,S0p,S0pp)
        S0pp_S1 = Transition(set,S0pp,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S1pp = Transition(set,S1p,S1pp)
        S1pp_S0 = Transition(set,S1pp,S0)
        self.transitions = set
 
        ## State labels
        self.known = ( (S1, S0), (S1, S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S0p | S0pp | S1, S1p | S1pp | S0), (S0p | S0pp | S1, S1p | S1pp | S0) )
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )
        
        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0_S0p, S1_S1p), (S0_S0p, S1_S1p) )
        self.produced = ( (S0pp_S1, S1pp_S0), (S0pp_S1, S1pp_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S0p | S0p_S0pp | S0pp_S1, S1_S1p | S1p_S1pp | S1pp_S0), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Three time-step IP block used in EWF
class COp3pma(Model):

    def Specify(self,base):

        self.base = base
        self.width = 9

        ## States
        set = []
        S0 = State(self.base,'000',set)
        S0p = State(self.base,'001',set)
        S0pp = State(self.base,'011',set)
        S1 = State(self.base,'111',set)
        S1p = State(self.base,'110',set)
        S1pp = State(self.base,'100',set)
        self.states = set
        self.smth = S1
        self.dual = self.smth.P
        
        ## Transitions
        set = []
        S0_S0 =  Transition(set,S0,S0)
        S0_S0p = Transition(set,S0,S0p)
        S0p_S0pp = Transition(set,S0p,S0pp)
        S0pp_S1 = Transition(set,S0pp,S1)
        S1_S1 =  Transition(set,S1,S1)
        S1_S1p = Transition(set,S1,S1p)
        S1p_S1pp = Transition(set,S1p,S1pp)
        S1pp_S0 = Transition(set,S1pp,S0)
        self.transitions = set
 
        ## State labels
        self.known = ( (S1 | S1p | S1pp, S0 | S0p | S0pp), )
        self.stored = ( (S1 | S1p | S1pp, S0 | S0p | S0pp), )
        self.accepted = ( (S0p | S0pp | S1, S1p | S1pp | S0), (S1, S0) )
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )
        
        ## Transition labels
        self.required = ( (S0_S0p, S1_S1p), (S0pp_S1, S1pp_S0) )
        self.produced = ( (S0pp_S1, S1pp_S0), )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), )
        self.busy = ( (S0_S0p, S1_S1p), (S0pp_S1, S1pp_S0) )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], ]
        self.ooutputs = ( [], )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

## Case machines are multivalue MA
class Case(Model):

    ## Overloaded constructor
    def __init__(self,parent,cases,extensionof=0,stpart=0,smpart=0):
        self.cases = cases
        Model.__init__(self,parent,stpart,smpart)

        ## Compute state bits needed to encode cases
        self.digits = log(self.cases)/log(2)
        temp = int(self.digits)
        if (self.digits == temp):
            self.digits = temp
        else:
            self.digits = temp + 1

        self.extensionof = extensionof

    def Specify(self,base):

        if not(self.extensionof == 0):
            self.base = self.extensionof.base
            self.cbase = base
            self.width = (3*self.digits)
        else:
            self.base = base
            self.cbase = base + 6
            self.width = 6 + (3*self.digits)

        ## States

        ## Nonvalue specific States
        set = []
        R0 = State(self.base,'00',set)
        K0 = State(self.base,'01',set)
        R1 = State(self.base,'11',set)
        K1 = State(self.base,'10',set)
        self.states = set

        ## Encode cases
        self.case = []
        for i in range(0,self.cases-1):
            encode = self.Encode(i)
            casest = State(self.cbase,encode)
            self.case.append(casest)

        ## Last case is the default and covers all remaining encodings
        encode = self.Encode(self.cases-1)
        casest = State(self.cbase,encode)            
        for i in range(self.cases,pow(2,self.digits)):
            encode = self.Encode(i)
            casest = casest | State(self.cbase,encode)
        self.case.append(casest)


        ## Transitions
        
        ## build hold transitions -- case value may not change
        set = []
        for case in self.case:
            Transition(set,case,case)
        self.hold = VecOr(set)


        ## build data-phase transitions
        set = []
        K1_R0 =  Transition(set,K1,R0)         ## Case change allowed -- must be allowed!
        K0_K0 =  K0.P & K0.N & self.hold       ## No case change allowed
        set.append( K0_K0 )
        K0_R1 =  Transition(set,K0,R1)         ## Case change allowed -- must be allowed!
        K1_K1 =  K1.P & K1.N & self.hold       ## No case change allowed
        set.append( K1_K1 )
        self.transitions = set

        ## build control-phase transitions
        set = []
        R0_K0 =  Transition(set,R0,K0)         ## Case change allowed -- must be allowed!
        set.append( K0_K0 )
        R1_K1 =  Transition(set,R1,K1)         ## Case change allowed -- must be allowed!
        set.append( R1_K1 )
        K0_K1 =  Transition(set,K0,K1)         ## Bypass
        K1_K0 =  Transition(set,K1,K0)         ## Bypass

        set.append( K0_K0 )
        set.append( K1_K1 )
        self.transitionsC = set
        
        ## State labels
        self.known = ( (K1, K0), )
        self.stored = ( (K1, K0), )
        self.accepted = ( (R1 | K1, R0 | K0), )
        self.resolve = R1 | R0                 ## No need to distinguish between sense for resolve label
        self.taskstart = ( (K0, K1), )
        self.taskfinal = ( (K1, K0), )
        
        ## Transition labels
        self.required = ( (K0_R1, K1_R0), )
        self.produced = ( (K0_R1, K1_R0), )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), )
        self.busy = ( (K0_R1, K1_R0), )
        self.bypass = ( (K0_K1 + R1_K1, K1_K0 + R0_K0), )
        self.strictbypass = ( (K0_K1, K1_K0), )

        ## Dynamic information
        self.memcare = [ [mgr.Zero(), mgr.Zero()], ]
        self.ooutputs = ( [], )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]
 
        ## The univabs is a single all-one state minterm of ROBDD variables encoding a value
        ## It is used during validation, etc.
        encode = self.Encode(pow(2,self.digits)-1)
        self.univabs = State(self.cbase,encode)

        ## Smooth includes all ROBDD variables used in a case MA
        self.smth = R1 & self.univabs
        self.dual = R1.P ## Don't flip guards


    ## Overloaded BuildTR for Case MA
    def BuildTR(self):

        ## Base component lives in one partition
        self.TR = Trans(1)

        ## Transitions
        self.TR[0].rel = VecOr(self.transitions)
        
        ## Noncase value ROBDD variables are smoothed in the first partition
        tsmth = State(-1,self.smth)
        tsmth.Smooth(self.univabs)
        self.TR[0].pssmth[0] = tsmth.P
        self.TR[0].nssmth[0] = tsmth.N
        self.TR[0].tagsmth[0] = tsmth.T
        
        ## Case value ROBDD variables are smoothed in the second partition
        self.TR[0].pssmth.append(self.univabs.P)
        self.TR[0].nssmth.append(self.univabs.N)
        self.TR[0].tagsmth.append(self.univabs.T)

        ## Maping
        self.TR.SetMapInfo(self.smth.P,self.smth.N,self.smth.T)


    ## Overloaded BuildcTR for Case MA -- Must be called after BuildTR
    def BuildcTR(self):

        ## Copy existing TR
        self.cTR = Trans(0,self.TR)

        ## Transitions
        self.cTR[0].rel = VecOr(self.transitionsC)

    ## Create a Boolean string encoding val
    def Encode(self,val):
        
        result = ""
        
        for i in range(0,self.digits):
            bit = val % 2;
            val = val >> 1;
            if bit:
                result = "1" + result
            else:
                result = "0" + result

        return result

    ## OnlyWhen functions define control blocks for MA in a composition
    
    ## malist contains ma that belong to this control block
    ## incaselist contains integers cases identifiying an MA's control block
    ## jointlist contains integer cases needing further case conditioning    
    def OnlyWhen(self,incaselist,malist, jointlist=() ):

        ## instate identifies when ma is in control block
        instate = State(-2)
        for i in incaselist:
            instate = instate | self.case[i]

        ## outstate identifies when ma is not in control block
        outstate = instate
        for i in jointlist:
            outstate = outstate | self.case[i]
        outstate = ~outstate

        ## update each ma in malist with appropriate control block info
        ## restricts by AND bmoth icb and ncb
        for ma in malist:
            
            ma.icb[0] = ma.icb[0] & (instate & self.syncedto.known[0][0])
            ma.icb[1] = ma.icb[1] & (instate & self.syncedto.known[0][1])

            ma.ncb[0] = ma.ncb[0] & (outstate & self.syncedto.known[0][0])
            ma.ncb[1] = ma.ncb[1] & (outstate & self.syncedto.known[0][1])

    ## no jointlist
    ## restricts icb but expands ncb
    def OnlyWhen1(self,incaselist,malist):

        instate = State(-2)
        for i in incaselist:
            instate = instate | self.case[i]

        outstate = ~instate
            
        for ma in malist:
            ma.icb[0] = ma.icb[0] & (instate & self.syncedto.known[0][0])
            ma.icb[1] = ma.icb[1] & (instate & self.syncedto.known[0][1])

            ma.ncb[0] = ma.ncb[0] | (outstate & self.syncedto.known[0][0])
            ma.ncb[1] = ma.ncb[1] | (outstate & self.syncedto.known[0][1])

    ## no jointlist
    ## refines control cases within a control case (control cases is already covered in icb)
    ## expands ncb and with removed cases from icb
    ## restricts icb
    def OnlyWhen2(self,incaselist,malist):

        instate = State(-2)
        for i in incaselist:
            instate = instate | self.case[i]

        outstate = ~instate
            
        for ma in malist:
            ## The outer exclude, now the inner exclude if the outer includes
            ma.ncb[0] = ma.ncb[0] | ( ma.icb[0] & self.syncedto.known[0][0] & outstate )
            ma.ncb[1] = ma.ncb[1] | ( ma.icb[1] & self.syncedto.known[0][1] & outstate )

            ## Now update to only include in inner case
            ma.icb[0] = ma.icb[0] & (instate & self.syncedto.known[0][0])
            ma.icb[1] = ma.icb[1] & (instate & self.syncedto.known[0][1])


    ## Overloaded for case
    ## Does not allow dependencies, capacities to or from case -- Use the synced MA!
    def BuildTRcTRConstraints(self, setTR, setcTR):

        ## Expression 4.19 in dissertation
        term0 = self.parent.sense.known[0][0].N & self.ncb[0].N & ~self.taskfinal[0][0].P
        term1 = self.parent.sense.known[0][1].N & self.ncb[1].N & ~self.taskfinal[0][1].P
        
        if (term0.Empty() or term1.Empty()):
            print "Term0 or term1 in a bypass construction is empty for",self.name

        ## If term0 or term1 is true, then must bypass to next sense task start states
        ## else any transition (including resolve) are allowed
        imp0 = ( term0 & self.taskstart[0][1].N) + ( ~term0 & ~self.strictbypass[0][0] )
        imp1 = ( term1 & self.taskstart[0][0].N) + ( ~term1 & ~self.strictbypass[0][1] )

        setcTR.append(imp0 & imp1)

    ## Synchronizes a particular valueless output operand production with a multivalue MA
    def Sync(self,set,opma,mai):
    
        set.append( opma.produced[mai][0] >> self.produced[0][0] )
        set.append( opma.produced[mai][1] >> self.produced[0][1] )
        set.append( self.produced[0][0] >> opma.produced[mai][0] )
        set.append( self.produced[0][1] >> opma.produced[mai][1] )

        self.syncedto = opma

    
## Rename is an experimental model used to simplify operand resolution
## It is another MA that collecs all pre operand resolution point
## And "renames" them as one
## To prevent paying a time penalty for renaming, it is built in the control phase
class Rename(Model):

    def Specify(self,base):

        ## State instantiation
        self.base = base
        self.width = 3
    
        set = []
        S0 = State(self.base,'0',set)
        S1 = State(self.base,'1',set)
        self.states = set

        self.smth = S1
        self.dual = self.smth.P
        
        ## Transition instantiation
        set = []
        S0_S0 = Transition(set,S0,S0)
        S1_S1 = Transition(set,S1,S1)
        self.transitions = list(set)

        S0_S1 = Transition(set,S0,S1)
        S1_S0 = Transition(set,S1,S0)
        ## Active transitions are in cTR
        self.transitionsC = set

        self.known = ( (S1, S0), (S1, S0) )
        self.stored = ( (S1, S0), (S1, S0) )
        self.accepted = ( (S1, S0), (S1, S0), (S1, S0) )
        self.taskstart = ( (S0, S1), )
        self.taskfinal = ( (S1, S0), )
        self.required = ( (S0_S1, S1_S0), (S0_S1, S1_S0), (S0_S1, S1_S0) )
        self.produced = ( (S0_S1, S1_S0), (S0_S1, S1_S0) )
        self.forget = ( (self.stored[0][0].P & ~self.stored[0][0].N, self.stored[0][1].P & ~self.stored[0][1].N), 
                        (self.stored[1][0].P & ~self.stored[1][0].N, self.stored[1][1].P & ~self.stored[1][1].N) )
        self.busy = ( (S0_S1, S1_S0), )

        self.memcare = [ [mgr.Zero(), mgr.Zero()], [mgr.Zero(), mgr.Zero()] ]
        self.ooutputs = ( [], [] )
        self.icb = [State(0), State(0)]
        self.ncb = [State(0), State(0)]

    ## Problem with sifting?
    def BuildSiftGroup(self):
        return

    ## Build the control TR -- Must be called after BuildTR
    def BuildcTR(self):

        ## Copy existing TR
        self.cTR = Trans(0,self.TR)

        ## Transitions
        hTR = VecOr(self.transitionsC)
        
        ## Task bypass labels
        self.bypass = ( ( ~self.taskstart[0][1].P & self.taskstart[0][1].N, ~self.taskstart[0][0].P & self.taskstart[0][0].N ), )

        ## Add forcing terms
        hTR = hTR + ( self.bypass[0][0] + self.bypass[0][1] )

        ## Substitue new relation
        self.cTR[0].rel = hTR

    ## Build dependency terms for Rename machine
    ## Overloaded for rename
    def BuildTRcTRConstraints(self, setTR, setcTR):

        ## guarded enable
        genable = [mgr.Zero(), mgr.Zero()] ## True guarded
        gcb = [mgr.Zero(), mgr.Zero()] ## guarded control blocks
        
        for ilist in self.oinputs:
          
            for dt in ilist:

                ## Sanity checks on dt structure
                if not (dt.acp == self):
                    print "WARNING: DT accepter is",dt.acp.name,"but built for",self.name

                if not (dt.ai ==0):
                    print "WARNING: DT accepter index is",dt.ai,"but built for",0

                ## Compute appropriate guards Intersection of producer, accepter and special conditions
                guard0 = self.icb[0].N & dt.prd.icb[0].N & dt.pcb[0].N
                guard1 = self.icb[1].N & dt.prd.icb[1].N & dt.pcb[1].N

                if guard0.Empty() or guard1.Empty():
                    print "WARNING: Guard is empty for operand resolution from",dt.prd.name,"to",self.name

                inside0 = guard0 & gcb[0]
                outside0 = guard0 & ~gcb[0] 

                inside1 = guard1 & gcb[1]
                outside1 = guard1 & ~gcb[1]

                if not inside0.Empty():
                    print "WARNING: Found overlapping control for",self.name,"input",0
                    ## Determine portion to constrain
                    toconstrain = genable[0] & inside0
 
                    ## remove toconstrain
                    genable[0] = genable[0] - toconstrain
 
                    ## add inside0 back
                    constrained = toconstrain & dt.prd.known[dt.pi][dt.si].P
                    genable[0] = genable[0] + constrained
 
                if not inside1.Empty():
 
                    ## Determine portion to constrain
                    toconstrain = genable[1] & inside1
 
                    ## remove toconstrain
                    genable[1] = genable[1] - toconstrain
 
                    ## add constrained back
                    constrained = toconstrain & dt.prd.known[dt.pi][abs(dt.si-1)].P
                    genable[1] = genable[1] + constrained

                ## Add outside terms
                genable[0] = genable[0] + (outside0 & dt.prd.known[dt.pi][dt.si].P)
                genable[1] = genable[1] + (outside1 & dt.prd.known[dt.pi][abs(dt.si-1)].P)
                gcb[0] = gcb[0] + outside0
                gcb[1] = gcb[1] + outside1

        ## Build force schedule constraints
        ## control block will not include self AND
        ## Control will be known AND
        ## parents will be known AND
        ## it is not already forced

        term0 = self.ncb[0].N & ~self.taskstart[0][1].P
        term1 = self.ncb[1].N & ~self.taskstart[0][0].P

        if (term0.Empty() or term1.Empty()):
            print "Term0 or term1 in a bypass construction is empty for",self.name

        if (genable[0].Empty() or genable[1].Empty()):
            print "Genable terms are empty for",self.name
        else:
            print "Operand resolution point for",self.name,"input 0"

        ## input edge implies dependencies are satisfied or forced
        if not (genable[0].Empty() & term0.Empty() ): 
            setcTR.append( self.required[0][0] >> (genable[0] | term0 ) )
        if not (genable[1].Empty() & term1.Empty() ):     
            setcTR.append( self.required[0][1] >> (genable[1] | term1 ) )

        ## force implies no idle
        if not (term0.Empty()):
           setcTR.append( (term0) >> ~(self.taskstart[0][0].P & self.taskstart[0][0].N) )
        if not (term1.Empty()):
           setcTR.append( (term1) >> ~(self.taskstart[0][1].P & self.taskstart[0][1].N) )










