## Copyright (C) 1998-2001 by
## Steve Haynal
## Department of Electrical and Computer Engineering,
## University of California, Santa Barbara.
##
##  Permission to use, copy, modify and/or distribute this software
##  and its documentation for any purpose is hereby granted without
##  fee, subject to the following terms and conditions:
##
##  1.  The above copyright notice and this permission notice must
##  appear in all copies of the software and related documentation.
##
##  2.  This software may not be called "PYABSS" or "ABSS" if it has
##  been modified in any way, without the specific prior permission
##  of Andrew Seawright.
##
##  3.  THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
##  EXPRESSED, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
##  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
##
##  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA OR THE AUTHORS OF THIS
##  SOFTWARE BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
##  CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING
##  FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
##  POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF
##  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
##
##  notes:
##
##  The development of this software was supported through grants from
##  Intel Inc., the University of California Micro Program and
##  National Science Foundation
##                                                        

##############################################################################################
## This file describes support routines, transition relation classes and path set classes    
## ABSS dissertation chapters 4 and 5, with emphasis on chapter 5, are implmented here.               
##
## Steve Haynal, 1/22/01
##
## This file is organized as follows:
##       Misc support routines (Some dependency, etc, Chapter 4)
##       ShiftBDD class used to shift BDDs (Chapter 4, iterates)
##       Group class used to build r-combination constraints ((Chapter 4)
##       DT class used to encapsulate a dependency tuple (Chapter 2)
##       State class encapsulates and associates ROBDD variables with present state, next state and tag state (Chapter 3,4,5)
##       TransPart class encapsulates a partition of a transition relation (Chapter 4,5)
##       Trans class encapsulates an entire partitioned transition relation (Chapter 4,5)
##       PathSet encapsulates a path set (Chapter 5)
##       Experimental code for reading blif files
##
## 

import pycudd
import time, re, sys
global mgr
mgr = pycudd.DdManager()
mgr.SetDefault()
from pycudd import DdArray



################################################################
## Commonly used functions

## Returns product of a list of ROBDDs
def VecAnd(args):
    vec = DdArray(len(args))
    for st in args:
        vec.Push(st)
    return vec.And()

## Returns sum of a list or ROBDDs
def VecOr(args):
    vec = DdArray(len(args))
    for st in args:
        vec.Push(st)
    return vec.Or()

## Converts a list or ROBDDs to a DdArray object
def List2Vec(args):
    vec = DdArray(len(args))
    for arg in args:
        vec.Push(arg)
    return vec

## Returns a ROBDD representing the transition(s)
## from state(s) in grp1 to state(s) in grp2
## grp1 and grp2 may be single states or lists of states
def Transition(set,grp1,grp2):

    try:
        vec = DdArray(len(grp1))
        for arg in grp1:
            vec.Push(arg.P)
        grp1 = vec.Or()
    except AttributeError:
        grp1 = grp1.P

    try:
        vec = DdArray(len(grp2))
        for arg in grp2:
            vec.Push(arg.N)
        grp2 = vec.Or()
    except AttributeError:
        grp2 = grp2.N

    result = grp1 & grp2

    set.append(result)
    return result


## Old-style dependency edge for DFG examples
def Enable(set,grp1,grp2):

    # Expecting state tuples by sense or single state
    try:
        # uses sense of sink
        minl = len(grp2)
        vec = DdArray(minl)

        for i in range(0,minl):
            try:
                vec.Push(grp2[i] >> grp1[i].P)
            except AttributeError:
                t = mgr.One()
                for src in grp1:
                    t = t & src[i].P
                vec.Push(grp2[i] >> t)

        set.append(vec.And())

    except TypeError:

        set.append(grp2 >> grp1.P)
        
def IEnable(set,grp1,grp2):

    # Expecting state tuple by sense
    minl = len(grp2)
    vec = DdArray(minl)

    for i in range(0,minl):
        try:
            vec.Push(grp2[i] >> grp1[(minl-1)-i].P)
        except AttributeError:
            t = mgr.One()
            for src in grp1:
                t = t & src[(minl-1)-i].P
            vec.Push(grp2[i] >> t)

    set.append(vec.And())

## Old-style capacity edge for DFG examples
def Capacity(set,grp1,grp2):

    # Expecting state tuple by sense
    minl = len(grp2)
    vec = DdArray(minl)

    for i in range(0,minl):
        try:
            vec.Push( (grp1[i].P & ~grp1[i].N) >> grp2[i].N )
        except AttributeError:
            t = mgr.Zero()
            for src in grp1:
                t = t + (src[i].P & ~src[i].N)
            vec.Push(t >> grp2[i].N)
            
    set.append(vec.And())

def ICapacity(set,grp1,grp2):

    # Expecting state tuple by sense
    minl = len(grp2)
    vec = DdArray(minl)

    for i in range(0,minl):
        try:
            vec.Push( (grp1[i].P & ~grp1[i].N) >> grp2[(minl-1)-i].N )
        except AttributeError:
            t = mgr.Zero()
            for src in grp1:
                t = t + (src[i].P & ~src[i].N)
            vec.Push(t >> grp2[(minl-1)-i].N)

    set.append(vec.And())

## Builds an implication between two ROBDDs and stores it in set
def Implication(set,grp1,grp2):

    set.append(grp1 >> grp2)

## Builds a double implication between two ROBDDs and stores it in set
def DoubleImplication(set,grp1,grp2):

    set.append(grp1 >> grp2)
    set.append(grp2 >> grp1)
 
## Imposes a partial order
## See FDCT benchmark example for information on use
def PartialOrder(set,grp1,grp2):

    # Expecting state tuples by sense or single state
    try:
        # uses sense of sink
        minl = len(grp2)
        vec = DdArray(minl)

        for i in range(0,minl):
            try:
                vec.Push(grp2[i] >> ( grp1[i].P | grp1[i].N))
            except AttributeError:
                t = mgr.One()
                for src in grp1:
                    t = t & ( src[i].P | src[i].N )
                vec.Push(grp2[i] >> t)

        set.append(vec.And())

    except TypeError:

        set.append(grp2 >> ( grp1.P | grp1.N) )
        
def IPartialOrder(set,grp1,grp2):

    # Expecting state tuple by sense
    minl = len(grp2)
    vec = DdArray(minl)

    for i in range(0,minl):
        try:
            vec.Push(grp2[i] >> (grp1[(minl-1)-i].P | grp1[(minl-1)-i].N)) 
        except AttributeError:
            t = mgr.One()
            for src in grp1:
                t = t & ( src[(minl-1)-i].P | src[(minl-1)-i].N )
            vec.Push(grp2[i] >> t)

    set.append(vec.And())



###################################################################
## A shift object allows a ROBDD to be shifted to a new range of
## node indices -- even if these ranges overlap
## It is used to build new MA and iterate instances from an original
## instance. Its use is best seen in the rotor and risc benchmark
## examples.


class ShiftBDD :
    ## lowi and highi define the range of ROBDD indices that must be shifted
    ## newlowi defines the new index (lowset) where the ROBDD should be shifted to 
    def __init__(self, lowi, highi, newlowi):

        overlap = (highi - newlowi)
        newhighi = (highi-lowi) + newlowi

        self.shftamt = newlowi - lowi

        self.svo = []
        self.svn = []

        print "Original span",lowi,highi,highi-lowi
        print "New span",newlowi,newhighi,newhighi-newlowi
        print "Overlap",overlap

        if overlap > 0:

            ## Move the non overlapping portion
            original = []
            for i in range(overlap,highi):
                original.append(mgr.IthVar(i))

            shifted = []
            for i in range(newlowi+overlap,newhighi):
                shifted.append(mgr.IthVar(i))

            veco = List2Vec(original)
            vecs = List2Vec(shifted)

            self.svo.append(veco)
            self.svn.append(vecs)

            ## Move the overlapping portion
            original = []
            for i in range(lowi,overlap):
                original.append(mgr.IthVar(i))

            shifted = []
            for i in range(newlowi,newlowi+overlap):
                shifted.append(mgr.IthVar(i))

            veco = List2Vec(original)
            vecs = List2Vec(shifted)

            self.svo.append(veco)
            self.svn.append(vecs)

        else:

             ## shift
            original = []
            for i in range(lowi,highi):
                original.append(mgr.IthVar(i))

            shifted = []
            for i in range(newlowi,newhighi):
                shifted.append(mgr.IthVar(i))

            veco = List2Vec(original)
            vecs = List2Vec(shifted)
            self.svo.append(veco)
            self.svn.append(vecs)

    ## Main routine called to shift a bdd to a new range
    def Shift(self,bdd):

        result = bdd
        for i in range(0,len(self.svo)):
            result = result.SwapVariables(self.svo[i], self.svn[i], len(self.svo[i]))

        return result


        
################################################################
## This class groups ROBDD terms to construct r-combination
## constraints

class Group :
    ## low and high are the r-combination bounds
    ## args is a list of ROBDDs
    ## for example low=0 high=2 will build
    ## "At most 2 ROBDDs from the arg list are true"
    ## likewise, low=4 high=4 will build
    ## "Exactly 4 ROBDDs from the arg list are true"
    def __init__(self, low, high, args):
        self.low = low
        self.high = high
        self.args = args
        self.constraint = mgr.One()

    ## Builds actual r-combination constraint
    ## only returns the constraint but will save it if save is true
    ## expects the self.args is a (odd, even) BDD tuple as in looping examples
    ## the self.args may be in just one sense (odd, )
    ## this routine is commonly used for busy-labeled transitions where sense is indistinguishable
    def BuildActive(self,save=0):

        vec = DdArray(len(self.args))
        for arg in self.args:
            ## Try for subterms
            try:
                temp = mgr.Zero()
                for iarg in arg:
                    temp = temp + iarg
            except AttributeError:
                print "Active was not by sense!"
                temp = arg

            vec.Push(temp)

        constraint = vec.Constraint(self.low,self.high)

        if save:
            self.constraint = constraint

        return constraint

    ## Builds resource type constraint over all senses for ORed groups
    ## Expects that self.args is a list of lists of tuples, i.e. ( ( (odd,even), (odd,even),), ... )
    ## Before, each item in self.args was an odd/even tuple of ROBDDs and r-combinations were determined
    ## With BuildActiveSetOR, each item in self.args is in fact a short list of odd/even tuple ROBDDs
    ## an r-combination requires that only ONE item (hence OR) in this inner list need be true to satisfy that
    ## this item (from the outer list) be counted in an r-combination constraint
    def BuildActiveSetOR(self,save=0):

        vec = DdArray(len(self.args))
        for argo in self.args:
            temp = mgr.Zero()
            for arg in argo:
                 ## Try for sense subterms
                try:
                    for iarg in arg:
                        temp = temp + iarg
                except AttributeError:
                    temp = temp + arg
                    
            vec.Push(temp)

        constraint = vec.Constraint(self.low,self.high)

        if save:
            self.constraint = constraint

        return constraint


###################################################################
## The DT class encapsulates a dependency tuple
class DT :

    ## Constructs a dt
    ## prdin is the producer in a dependency tuple
    ## piin is the producer's output that is expected
    ## siin is the sense that is expected
    ## cbin defaults to the producer's control block if cbin=0, else a new control-block may be specified
    ## For example DT(task5,3,0)
    ## builds a dependency tuple that requires output 3 from task5 in a normal (inter-iteration) sense
    ## It is valid under all control-blocks where task5 is valid (i.e. no control-block override)
    def __init__(self, prdin, piin, siin, cbin=0):

        ## Producer
        self.prd = prdin

        ## Producer's output index
        self.pi = piin

        ## Sense index
        self.si = siin

        if cbin == 0:
            self.pcb = self.prd.icb
        else:
            self.pcb = cbin

    ## Adds an accepter to the dependency tuple
    ## acpin is the accepting task (an MA)
    ## aiin is the input index for the accepting task that depends on the producer's result
    ## cbin is a control-block override
    def AddAccepter(self, acpin, aiin, cbin=0):

        self.acp = acpin

        self.ai = aiin

        if cbin == 0:
            self.acb = self.acp.icb
        else:
            self.acb = cbin

    ## dumps a dot format edge line for a dependency tuple
    def DotEdgeLine(self, outfile):

        str = '\t"' + self.prd.name + '":o' + repr(self.pi) + ' -> "' + self.acp.name + '":i' + repr(self.ai) + ';\n'
        outfile.write(str)


################################################################
# The state class encapsulates a present state, next state and tag ROBDDs
class State :
    # Constructs a state
    def __init__(self, base, arg='-', set=0):

        ## Copy state, arg must be state
        if base== -1:
            self.T = arg.T
            self.P = arg.P
            self.N = arg.N
            self.base = arg.base

        ## Build an empty state
        elif base== -2:
            self.T = mgr.Zero()
            self.P = mgr.Zero()
            self.N = mgr.Zero()
            self.base = 0
            
        ## Encode state, arg must be string (0,1,-)
        else:
            self.T = mgr.StateCube(arg,base,2,3)
            self.P = mgr.StateCube(arg,base,1,3)
            self.N = mgr.StateCube(arg,base,0,3)
            self.base = base

        ## Try and save in set
        try:
            set.append(self)
        except AttributeError:
            pass
        
    # String representation
    def __repr__(self):
        return "<State Instance>"
    
    # States may be ORed together
    def __or__(self,arg0):
        val = State(-1,self)
        val.T = self.T | arg0.T
        val.P = self.P | arg0.P
        val.N = self.N | arg0.N
        return val

    # States may be ANDed together
    def __and__(self,arg0):
        val = State(-1,self)
        val.T = self.T & arg0.T
        val.P = self.P & arg0.P
        val.N = self.N & arg0.N
        return val

    # States may be inverted
    def __invert__(self):
        val = State(-1,self)
        val.T = ~self.T
        val.P = ~self.P
        val.N = ~self.N
        return val

    # States may be shifted
    # shft is a shift object
    def Shift(self, shft):

        ## Shift relation
        self.T = shft.Shift(self.T)
        self.P = shft.Shift(self.P)
        self.N = shft.Shift(self.N)
        self.base = self.base + shft.shftamt

    # States may be smoothed (existentially quantified)
    # by another state
    def Smooth(self, other):
        ## Shift relation
        self.T = self.T.ExistAbstract(other.T)
        self.P = self.P.ExistAbstract(other.P)
        self.N = self.N.ExistAbstract(other.N)
 

##############################################################
# Transition relations are partitioned
# TransPart encapsulates a particular transition partition
#
# Important data members are:
#
#    self.rel is the ROBDD relation
#    self.pssmth is a list of single minterm ROBDDs describing which variables are in the ps (present
#    state) position. There is a list because smoothing is done in steps. i.e. the guard variables
#    are not immediately smoothed. For example,
#    [---11--,-----11] would first smooth the ---11-- ROBDD variables and later the -----11 ROBDD variables.
#    self.nssmth and self.tagsmth similar to self.pssmth except covering ns and tag variables

class TransPart :
    def __init__(self,orig=0):
        ## Do not copy original
        if orig==0:
            self.rel = mgr.One()
            self.pssmth = [mgr.One(),]
            self.nssmth = [mgr.One(),]
            self.tagsmth = [mgr.One(),]
        ## Make a copy of orig
        else:
            self.rel = orig.rel
            self.pssmth = list(orig.pssmth)
            self.nssmth = list(orig.nssmth)
            self.tagsmth = list(orig.tagsmth)
           
    def Info(self):
        print 'Nodes in relation is',self.rel.DagSize()

    ## Collapse builds product of several partitions
    ## This is used to collapse a partitioned transition relation into
    ## a nonpartitioned one
    def Collapse(self,other):
        self.rel = self.rel & other.rel
    
        for i in range(0,len(self.pssmth)):
            try:
                self.pssmth[i] = self.pssmth[i] & other.pssmth[i]
            except IndexError:
                pass
            
        for i in range(0,len(self.nssmth)):
            try:
                self.nssmth[i] = self.nssmth[i] & other.nssmth[i]             
            except IndexError:
                pass

        for i in range(0,len(self.tagsmth)):
            try:
                self.tagsmth[i] = self.tagsmth[i] & other.tagsmth[i]             
            except IndexError:
                pass

    ## Entirely smooth a term (Exististentially abstract all ps, ns and tag variables in this partition)
    def Smooth(self,term):

        self.rel = self.rel.ExistAbstract(term)

        for i in range(0,len(self.pssmth)):
            self.pssmth[i] = self.pssmth[i].ExistAbstract(term)

        for i in range(0,len(self.nssmth)):
            self.nssmth[i] = self.nssmth[i].ExistAbstract(term)

        for i in range(0,len(self.tagsmth)):
            self.tagsmth[i] = self.tagsmth[i].ExistAbstract(term)

    ## Shift a partition
    def Shift(self, shft):

        ## Shift relation
        self.rel = shft.Shift(self.rel)

        for i in range(0,len(self.pssmth)):
            self.pssmth[i] = shft.Shift(self.pssmth[i])

        for i in range(0,len(self.nssmth)):
            self.nssmth[i] = shft.Shift(self.nssmth[i])
            
        for i in range(0,len(self.tagsmth)):
            self.tagsmth[i] = shft.Shift(self.tagsmth[i])
            
##############################################################
# This class is a partitioned transition relation
# Top-level transition relation

class Trans :
    def __init__(self,parts,orig=0):

        ## Create a new Trans object
        if orig==0:
            ## Number of partitions and list of partitions
            self.parts = parts
            self.partitions = []
        
            for j in range(0,self.parts):
                self.partitions.append(TransPart())

            ## Attributes assigned dynamacially to establish proper mapping
            self.ps = mgr.One()
            self.ns = mgr.One()
            self.tag = mgr.One()
            ## self.vecps DdArray of present state variables
            ## self.vecns DdArray of next state variables

        ## Copy a Trans object
        else:
            self.parts = orig.parts
            self.partitions = []

            self.ps = orig.ps
            self.vecps = orig.vecps
            self.ns = orig.ns
            self.vecns = orig.vecns
            self.tag = orig.tag
            self.vectag = orig.vectag

            for part in orig.partitions:
                self.partitions.append(TransPart(part))

    ## Add a partition to this Trans object 
    def AddTransPart(self,tpart):
        self.partitions.append(tpart)
        self.parts = len(self.partitions)

    ## Called when constructing a Trans object for a composite task
    ## Information is collected in lists and then later processed to
    ## create the desired Trans object
    def PreCompose(self):
        self.ps = []
        self.ns = []
        self.tag = []
        for part in self.partitions:
            part.rel = []
            part.pssmth[0] = []
            part.nssmth[0] = []
            part.tagsmth[0] = []

    ## Used to collect information from a composite-member MA during
    ## construction of a composite task
    ## stpart is the base partition that the submodel state partitions should start at
    ## smpart is the base inner smooth step that the submodel should start at
    def Add(self,other,stpart=0,smpart=0):

        self.ps.append(other.ps)
        self.ns.append(other.ns)
        self.tag.append(other.tag)

        while ((len(other.partitions)+stpart) > len(self.partitions)):
            TR = TransPart()
            TR.rel = []
            TR.pssmth[0] = []
            TR.nssmth[0] = []
            TR.tagsmth[0] = []
            self.partitions.append(TR)
            
        for i in range(0,len(other.partitions)):

            self[i+stpart].rel.append(other[i].rel)

            while ((len(other[i].pssmth)+smpart) > len(self[i].pssmth)):
                self[i+stpart].pssmth.append([])

            for j in range(0,len(other[i].pssmth)):
                self[i+stpart].pssmth[j+smpart].append(other[i].pssmth[j])
                
            while ((len(other[i].nssmth)+smpart) > len(self[i].nssmth)):
                self[i+stpart].nssmth.append([])

            for j in range(0,len(other[i].nssmth)):
                self[i+stpart].nssmth[j+smpart].append(other[i].nssmth[j])

            while ((len(other[i].tagsmth)+smpart) > len(self[i].tagsmth)):
                self[i+stpart].tagsmth.append([])

            for j in range(0,len(other[i].tagsmth)):
                self[i+stpart].tagsmth[j+smpart].append(other[i].tagsmth[j])

    ## Reduces all the collected information (in lists) to actual structures
    ## Final step of constructing a composite task
    def Compose(self):

        vec = List2Vec(self.ps)
        self.ps = vec.And()
        self.vecps = self.ps.Vector()
       
        vec = List2Vec(self.ns)
        self.ns = vec.And()
        self.vecns = self.ns.Vector()

        vec = List2Vec(self.tag)
        self.tag = vec.And()
        self.vectag = self.tag.Vector()
        
        for part in self.partitions:
            vec = List2Vec(part.rel)
            part.rel = vec.And()

            for i in range(0,len(part.pssmth)):
                vec = List2Vec(part.pssmth[i])
                part.pssmth[i] = vec.And()

            for i in range(0,len(part.nssmth)):
                vec = List2Vec(part.nssmth[i])
                part.nssmth[i] = vec.And()

            for i in range(0,len(part.tagsmth)):
                vec = List2Vec(part.tagsmth[i])
                part.tagsmth[i] = vec.And()  

    ## Reduce TR to nonpartitioned
    ## Everything is put in partition 0
    def Collapse(self):

        for i in range(1,len(self.partitions)):
            self.partitions[0].Collapse(self.partitions[i])
         
        self.partitions = [ self.partitions[0], ]
        self.parts = 1

    ## Entirely smoth variables from term, all partitions and all ROBDD variables
    def Smooth(self,term):

        for i in range(0,len(self.partitions)):
            self.partitions[i].Smooth(term)


    ## Compute an image of ts
    ## if no partition slice is specified, all partitions are used in the computation
    ## if no smooth slice is specified, all smooth partitions within a TransPart are used
    def Image(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        
        result = ts

        ## Compute image over requested partition slice
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            
            result = result.AndAbstract(self[i].rel,self[i].pssmth[ismthslice[0]])
         
            ## Continue smoothing if multiple smooth cubes
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].pssmth[j])
          
        result = result.VarMap()
      
        return result
    
    def PreImage(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts.VarMap()

        ## Compute preimage over requested partition slice
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.AndAbstract(self[i].rel,self[i].nssmth[ismthslice[0]])

            ## Continue smoothing if multiple smooth cubes
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].nssmth[j])

        return result

    ## Intersect ts with the transition relation and smooth ps variables
    def AndSmoothPS(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.AndAbstract(self[i].rel,self[i].pssmth[ismthslice[0]])
            
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].pssmth[j])

        return result

    def AndSmoothNS(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.AndAbstract(self[i].rel,self[i].nssmth[ismthslice[0]])
            
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].nssmth[j])

        return result

    ## smooth ps variables from ts
    def SmoothPS(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.ExistAbstract(self[i].pssmth[ismthslice[0]])
            
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].pssmth[j])

        return result
    
    def SmoothNS(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.ExistAbstract(self[i].nssmth[ismthslice[0]])
            
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].nssmth[j])

        return result

    ## smooth tag variables from ts
    def SmoothTS(self, ts, partslice=(0,0), smthslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            if smthslice[0] == smthslice[1]: ismthslice = (0,len(self[i].pssmth))
            else: ismthslice = smthslice
            result = result.ExistAbstract(self[i].tagsmth[ismthslice[0]])
            
            for j in range(ismthslice[0]+1,ismthslice[1]):
                result = result.ExistAbstract(self[i].tagsmth[j])

        return result

    ## intersect ts with the transition relation
    def Intersect(self, ts, partslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        result = ts
             
        for i in range(partslice[0],partslice[1]):
            result = result & self[i].rel
     
        return result

    # Greatest fixed point
    def GreatestFP(self, term,  partslice=(0,0), smthslice=(0,0)):
        rs = mgr.Zero()
        cs = term

        i = 1
        while (cs != (cs & rs)):
            rs = rs + cs
            cs = self.Image(cs, partslice, smthslice)
            print "GFP" + repr(i),
            i = i + 1

        return rs

    # Least fixed point
    # FIXME: Modified for clock phases!
    def LeastFP(self, term, partslice=(0,0), smthslice=(0,0)):
        lcs = mgr.One()
        cs = term

        i = 1
        while (lcs != cs):
            lcs = cs
            cs = self.Image(cs,partslice,smthslice)
            cs = cs & lcs
            print "LFP" + repr(i),
            i = i + 1

        return cs

    ## Prune relations in the partition slices specified
    ## to contain states only in pruneps
    ## prunes both present-state and next-state positions of the TR
    def PruneS(self, pruneps, partslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))
        prunens = pruneps.VarMap()

        for i in range(partslice[0],partslice[1]):
            self[i].rel = self[i].rel & pruneps
            self[i].rel = self[i].rel & prunens

    ## Prune relations by prunet (using AND) in the partition slices specified
    def PruneT(self, prunet, partslice=(0,0)):
        if partslice[0] == partslice[1]: partslice = (0,len(self.partitions))

        for i in range(partslice[0],partslice[1]):
            self[i].rel = self[i].rel & prunet

    ## set ROBDD map ps-ns, tags remain in same dspot
    def SetVarMap(self):
        self.vecps.SetVarMap(self.vecns)

    ## cellect information to build a map
    def SetMapInfo(self,ps,ns,tag):
        self.ps = ps
        self.ns = ns
        self.tag = tag
        self.vecps = self.ps.Vector()
        self.vecns = self.ns.Vector()
        self.vectag = self.tag.Vector()

    # Prints transition relation information
    def Info(self):
        print 'Nodes in relations',
        for part in self.partitions:
            print part.rel.DagSize(),
        print

    # ReachabilityPrune
    # Starts with list of states in sslist
    # prunes the transition relation to the reachable state set from sslist
    def ReachabilityPrune(self,sslist,pruneslice=(0,0),partslice=(0,0),smthslice=(0,0)):
        start = time.clock()

        term = mgr.Zero()
        for arg in sslist:
	    term = term | arg.P

        cs = self.GreatestFP(term,partslice,smthslice)
        self.PruneS(cs,pruneslice)

        print
        print 'Reachability Prune in',time.clock()-start,'seconds.'
        print 'Peak Nodes',mgr.ReadPeakNodeCount()

    ## Shift a Trans object
    def Shift(self,shft):

        ## Shift all partitions
        for part in self.partitions:
            part.Shift(shft)

        ## Shift ps, ns, tag
        self.ps = shft.Shift(self.ps)
        self.ns = shft.Shift(self.ns)
        self.tag = shft.Shift(self.tag)

    ## Return a particualr partition
    def __getitem__(self,i):
        return self.partitions[i]


##############################################################################
## This class corresponds to the PathSet described in the ABSS dissertation
##
##   Member Data:
##     CMA is associated composite ma
##     tslist is list of time-step state sets (path set)
##     TRlist is list of time-step transition sets (transition set)
##     TR is the default transition relation, usually from CMA
##     cTR is the default control phase transition relation, usually from CMA
##     tslist[0] is start state or behavior cut
##     fs is expected final state or behavior cut, in all cases this is ts[0]~
##     rfs is reached final states
##
##   Member Functions:
##     FE is Forward Exploration
##       variations at least one final state reached, all states reached, a set final states covering all control cases reached, plus one step
##     BP is Backward Pruning
##       variations validation, termination accumulation

class PathSet :
    def __init__(self,CMA,Sbc,TR=0,cTR=0,orig=0):
        if orig==0:

            # Preserve reference to CMA
            self.CMA = CMA

            # Set time-step 0 to Sbc
            self.tslist = [Sbc,]

            # Compute expected fs
            self.fs = self.CMA.Dual(Sbc)

            # No fs states have been reached 
            self.rfs = mgr.Zero()

            # Default Copy of TR -- separated from CMA to support dynamic change
            if TR:
                self.TR = TR
            else:
                self.TR = Trans(0,self.CMA.TR)

            if cTR:
                self.cTR = cTR
            else:
                try:
                    self.cTR = Trans(0,self.CMA.cTR)
                except AttributeError:
                    self.cTR = 0
                    
            # TR list
            self.TRlist = []

            # Particular trace
            self.trace = []

        # Make a copy
        else:
            # CMA is a reference, not a copy
            self.CMA = orig.CMA

            # All ROBDDs are copies
            self.fs = orig.fs
            self.rfs = orig.rfs
            self.tslist = list(orig.tslist)
            self.trace = list(orig.trace)

            # Copy original TR
            self.TR = Trans(0,orig.TR)
            if orig.cTR:
                self.cTR = Trans(0,orig.cTR)
            else:
                self.cTR = 0
            
            # This copies TRlist items as new instances
            # TR in the TRlist may be changed dynamically
            self.TRlist = []
            tmap = {}
            tmap[orig.TR] = self.TR     
            for TR in orig.TRlist:
                try:
                    self.TRlist.append(tmap[TR])
                except KeyError:
                    ## Make and record copy
                    tmap[TR] = Trans(0,TR)
                    self.TRlist.append(tmap[TR])


    def __getitem__(self,i):
        return self.tslist[i]

    def __setitem__(self,i,val):
        self.tslist[i] = val

    def last(self):
        return len(self.tslist)-1

    # Check if term exists in tslist
    def FindTerms(self,term):

        i = 0
        for ts in self.tslist:
            if not (ts.Leq(~term)):
                print "Term exists in time-step set",i
            i = i + 1

    # Determines the largest ROBDD in the tslist
    def MaxSize(self):

        size = 0
        for ts in self.tslist:
            a = ts.DagSize()
            if a > size:
                size = a

        print "Max set size is",size,"nodes"

    def PruneForward(self, partslice=(0,0), smthslice=(0,0),aterm=0):

        ## rfs starts empty
        self.rfs = mgr.Zero()
            
        for i in range(0,self.last()):

            result = self.StepForward(i,0,partslice, smthslice, self[i+1])

            ## Accumulate termination
            if aterm:
                temp = aterm & result
                self.rfs = self.rfs + temp

            if result.Empty():
                print "FBail" + repr(i)
                return 0
            else:
                print "PF" + repr(i),
        return 1
    
    def PruneBackward(self, partslice=(0,0), smthslice=(0,0), aterm=0):

        ## rfs includes only TSi, other terminations will be accumulated
        self.rfs = self[self.last()]

        for i in range(self.last(),0,-1):

            if aterm:
                temp = self[i-1] & aterm ## Preserve term
             
            result = self.StepBackward(i,partslice, smthslice, self[i-1])

            if aterm:
                result = result + temp ## Preserve terms
                self[i-1] = result
                ## Accumulate terminations
                self.rfs = self.rfs + temp
                           
            if result.Empty():
                print "BBail" + repr(i)
                return 0
            else:
                print "PB" + repr(i),         
        return 1      
    
    ## Regular step forward
    def StepForward(self, i, TR=0, partslice=(0,0), smthslice=(0,0), pterm=0):

        ## TSi better exist!
        try:
            rs = self.TRlist[i].Image(self[i],partslice,smthslice)
        except IndexError:
            ## Substitute TR
            if TR:
                self.TRlist.append(TR)
            ## Default TR 
            else:
                self.TRlist.append(self.TR)
            rs = self.TRlist[i].Image(self[i],partslice,smthslice)

        if (pterm):
            rs = rs & pterm
  
        try:
            ## Replace existing TS
            self.tslist[i+1] = rs
        except IndexError:
            ## Add TS
            self.tslist.append(rs)
        return rs

    ## Regular step backward
    def StepBackward(self, i, partslice=(0,0), smthslice=(0,0), pterm=0):

        ## TS at i better exist! TRlist at i-1 better exist!
        rs = self.TRlist[i-1].PreImage(self.tslist[i],partslice,smthslice)
     
        ## Prune
        if (pterm): 
            rs = rs & pterm
          
        ## TS at i-1 better exist!
        self.tslist[i-1] = rs

        return rs

    ## Forward Exploration
    ## Early term provides a precheck for termination
    ## -1 just extends by one step
    def ForwardExploration(self, i=0, earlyterm=0):

        if i == -1:
            i = self.last()
            self.StepForward(i,0,(0,0),(0,1))
            i = i + 1

        ## Start at TS0 if i=0
        while 1:
            ## Check for possible termination
            temp = earlyterm.Leq(~self[i])            
            if not (temp):
                ## Check for true termination
                temp = self.fs.Leq(~self[i])
                if not (temp):
                    ## True termination
                    print "FTS" + repr(i),
                    ## Expand rfs
                    self.rfs = self.rfs + (self.fs & self[i])
                    break
                else:
                    ## Cut has toggled sense
                    print "KTS" + repr(i),
            else:
                ## No termination
                print "TS" + repr(i),

            ## Do regular forward step
            self.StepForward(i,0,(0,0),(0,1)) # Don't smooth tags
            i = i + 1

    ## Closure 
    def Closure(self,userfs=1):

        # wset is a copy
        wset = PathSet(0,0,0,0,self)

        # Closure fixed-point
        i = 0
        while 1:
            ## first == rfs
            wset[0] = wset[0] & wset.CMA.Dual(wset.rfs)
            wset[wset.last()] = wset[wset.last()] & wset.CMA.Dual(wset[0])

            ## Preserve reached fs during backward pruning
            if userfs:
                pterm = wset.rfs
            else:
                pterm = 0

            if not wset.PruneBackward((0,0),(0,1),pterm):
                print "Closure Failure" + repr(i)
                return 0
            print "BP" + repr(i),

            if (wset[0] == wset.CMA.Dual(wset.rfs)):
                print "Closure" + repr(i)
                return wset

            i = i + 1

            ## first == rfs
            wset[0] = wset[0] & wset.CMA.Dual(wset.rfs)
            wset[wset.last()] = wset[wset.last()] & wset.CMA.Dual(wset[0])

            if not wset.PruneForward((0,0),(0,1),wset.CMA.Dual(wset[0])):
                print "Closure Failure" + repr(i)
                return 0
            print "BP" + repr(i),

            if (wset[0] == wset.CMA.Dual(wset.rfs)):
                print "Closure" + repr(i)
                return wset

            i = i + 1

    ## Full Forward Exploration
    ## Early term provides a precheck for termination 
    def FullForwardExploration(self):

        rs = mgr.Zero()
        nrs = self[0]
        self.rfs = mgr.Zero()

        i=0
        while not (nrs == rs):

            rs = nrs

            print "TR step",
            self.StepForward(i,self.TR)
            i = i + 1
            print self[i].DagSize()

            if self.cTR:
                print "cTR step",
                self.StepForward(i,self.cTR)
                i = i + 1
                print self[i].DagSize()
                
            ## Expand rfs
            eterm =  (self.fs & self[i])
            if not eterm.Empty():
                self.rfs = self.rfs + eterm
                print "Expanded rfs at",i
                
            nrs = rs + self[i]

            if (nrs == rs):
                print "Terminated at",i
            else:
                print "Completed",i


    ## This assume a full path set with state sets control prioritized pruned
    def ValidatedClosure(self,validate=1):

        # wset is a copy
        wset = PathSet(0,0,0,0,self)

        # Closure fixed-point
        i = 0
        print "Starting outer fixed-point"
        while 1:
            ## first == rfs
            wset[0] = wset[0] & wset.CMA.Dual(wset.rfs)
            wset[wset.last()] = wset[wset.last()] & wset.CMA.Dual(wset[0])

            ## Preserve reached fs during backward pruning
            #pterm = wset.rfs
            pterm = wset.CMA.Dual(wset[0])

            ## Check for all paths
            if validate:
                if not wset.CMA.AllPathsCheck(wset[0]):
                    return 0
 
            j = wset.last()
            wset.rfs = wset[j]

            print "Size info",wset[0].DagSize(),wset[j].DagSize(),pterm.DagSize(),wset.rfs.DagSize()

            print "Starting inner fixed-point"
            while j>0:
                print "Vb" + repr(j),
                ## Preserve a possible lpfs, temp is a time step specific lpfs
                if j%2:
                    temp = wset[j-1] & pterm
                cs = wset.StepBackward(j, (0,0), (0,1), wset[j-1]) # Don't smooth guards
                ## Include expanded terms
                if j%2:
                    cs = cs + temp
                    wset.rfs = wset.rfs + temp
                print ".",

                ## Only validate if in known resolve phase
                if not j%2:

                    if validate:
                        # divide into two sets, nocond has no conditions resolving, conds has one or more conditions resolving
                        conds = cs & wset.CMA.resolve.P
                   
                        if not conds.Empty():

                              #Validate the set
                              ## Ignore tags during validation.
                              ##conds = wset.TRlist[j-1].SmoothNS(conds,(0,0),(2,3))
                              conds = wset.TRlist[j-1].SmoothTS(conds,(0,0),(0,1))
                              pconds = wset.CMA.Validate(conds)
                              
                              if conds != pconds:

                                  # Pruning occurred
                                  pconds = wset.TRlist[j-1].SmoothNS(pconds,(0,0),(1,2))
                                  noconds = cs & ~wset.CMA.resolve.P
                                  noconds = wset.TRlist[j-1].SmoothNS(noconds,(0,0),(1,2))
                                  cs = pconds + noconds
                              else:
                                  # Pruning did not occur
                                  cs = wset.TRlist[j-1].SmoothNS(cs,(0,0),(1,2))
                        else:
                            # Partitioning and pruning did not occur
                            cs = wset.TRlist[j-1].SmoothNS(cs,(0,0),(1,2))
                            
                    ## No validation
                    else:
                        cs = wset.TRlist[j-1].SmoothNS(cs,(0,0),(1,2))
                else:
                    # Partitioning and pruning did not occur
                    cs = wset.TRlist[j-1].SmoothNS(cs,(0,0),(1,2))
                
                j = j-1
                print cs.DagSize()
                wset[j] = cs

            # Check for validation failure
            if wset[0].Empty():
                print 'VF'+repr(j)
                return 0

            print "Size info",wset[0].DagSize(),wset[j].DagSize(),pterm.DagSize(),wset.rfs.DagSize()

            # Check for closure
            if (wset[0] == wset.CMA.Dual(wset.rfs)):
                print "Closure" + repr(i)
                return wset
            else:
                ttt = ~(wset[0]) & (wset.CMA.Dual(wset.rfs))
                print "Not Closed" + repr(i)
            
            i = i + 1

           
    ## FIXME: Only prunes default TR not list
    def PruneS(self, pruneps, partslice=(0,0)):

        self.TR.PruneS(pruneps,partslice)

        for TR in self.TRlist:
            if not TR == self.TR:
                print "Different TRs in PathSet TRlist!"

        print "PathSet PruneS Done"

    def SmoothTimeSteps(self, smthterm):

        for i in range(0,len(self.tslist)):
            self.tslist[i] = self.tslist[i].ExistAbstract(smthterm)

        self.rfs = self.rfs.ExistAbstract(smthterm)

    #############################################
    ## Code for printing a witness schedule

    ## Completes trace structure
    def Trace(self,islice=(0,0),pterm=( (-1,0),)):

        if islice[0] == islice[1]: islice = (0,len(self.TRlist))

        cs = self[islice[0]]

        ptermi = 0
        for i in range(islice[0],islice[1]):
            ttr = self.TRlist[i].Intersect(cs)
            ttr = ttr & self[i+1].VarMap()

            ## Additional pterm constraints
            try:
                if (pterm[ptermi][0] == i) or (pterm[ptermi][0] == '*'):
                    print "Applying pterm to",i,"-",i+1
                    for j in range(1,len(pterm[ptermi])):
                        ttr = ttr & pterm[ptermi][j]
                    if (pterm[ptermi][0] == i):   
                        ptermi = ptermi + 1
            except IndexError:
                pass

            if ttr.Empty():
                print "TTR is empty at ",i
                return
            
            ttr = self.CMA.vecAll.PickOneMinterm(ttr)

            try:
                self.trace[i] = ttr
            except IndexError:
                self.trace.append(ttr)

            cs = self.TRlist[i].SmoothPS(ttr)
            cs = cs.VarMap()

    ## Prints trace
    def PrintTrace(self):

        i = 1
        for ttr in self.trace:
            print
            print "Time step",i,":"
            self.CMA.PrintTTR(ttr)
            i = i + 1

    def PrintStep(self,bm,nterm,pterm):

        ## assumes self.step is step to print
        ## assumes self.front is state to propogate

        ttr = self.TRlist[self.step].Intersect(self.front)
        ttr = ttr & self[self.step+1].VarMap()

        ttr = ttr & pterm

        ttr = bm.vecAll.PickOneMinterm(ttr)

        self.front = self.TRlist[self.step].SmoothPS(ttr)
        self.front = self.TRlist[self.step].SmoothTS(self.front,(0,0),(0,1))
        self.front = self.front.VarMap()

        ## This is an even step, so print activity
        print "Time-step",self.step
        bm.PrintTTR2(ttr)

        self.step = self.step + 1

        ## Propogate second step
        ttr = self.TRlist[self.step].Intersect(self.front)
        ttr = ttr & self[self.step+1].VarMap()

        ttr = ttr & nterm

        ttr = bm.vecAll.PickOneMinterm(ttr)

        self.front = self.TRlist[self.step].SmoothPS(ttr)
        self.front = self.TRlist[self.step].SmoothTS(self.front,(0,0),(0,1))
        self.front = self.front.VarMap()

        ## If lpfs state is reached, reset to time step 1
        temp = self[0] & bm.CMA.Dual(self.front)
        if not temp.Empty():
            print "Reached Sbc dual"
            self.front = bm.CMA.Dual(self.front)
            self.step = 0
        else:
            self.step = self.step + 1

    def Truncate(self,n):

        self.tslist = self.tslist[:n+1]
        self.TRlist = self.TRlist[:n]
            

















